from logging import Logger
from types import TracebackType
from typing import Optional
from typing import Text
from typing import Type

from qbittorrentapi._types import KwargsT
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.request import Request

logger: Logger

class Authorization(ClientCache):
    @property
    def is_logged_in(self) -> bool: ...
    def log_in(
        self,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    def log_out(self, **kwargs: KwargsT) -> None: ...

class AuthAPIMixIn(Request):
    @property
    def auth(self) -> Authorization: ...
    @property
    def authorization(self) -> Authorization: ...
    @property
    def is_logged_in(self) -> bool: ...
    def auth_log_in(
        self,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    @property
    def _SID(self) -> Optional[Text]: ...
    def _session_cookie(self, cookie_name: Text = "SID") -> Optional[Text]: ...
    def auth_log_out(self, **kwargs: KwargsT) -> None: ...
    def __enter__(self) -> "AuthAPIMixIn": ...
    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> bool: ...



from logging import Logger
from typing import Mapping
from typing import Optional
from typing import Text

from qbittorrentapi._types import JsonDictionaryT
from qbittorrentapi._types import JsonValueT
from qbittorrentapi._types import KwargsT
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.request import Request

logger: Logger

class ApplicationPreferencesDictionary(JsonDictionaryT): ...
class BuildInfoDictionary(JsonDictionaryT): ...

class Application(ClientCache):
    @property
    def version(self) -> Text: ...
    @property
    def web_api_version(self) -> Text: ...
    @property
    def webapiVersion(self) -> Text: ...
    @property
    def build_info(self) -> BuildInfoDictionary: ...
    @property
    def buildInfo(self) -> BuildInfoDictionary: ...
    def shutdown(self) -> None: ...
    @property
    def preferences(self) -> ApplicationPreferencesDictionary: ...
    @preferences.setter
    def preferences(self, value: Mapping[Text, JsonValueT]) -> None: ...
    def set_preferences(
        self,
        prefs: Optional[Mapping[Text, JsonValueT]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setPreferences = set_preferences
    @property
    def default_save_path(self) -> Text: ...
    @property
    def defaultSavePath(self) -> Text: ...

class AppAPIMixIn(Request):
    @property
    def app(self) -> Application: ...
    @property
    def application(self) -> Application: ...
    def app_version(self, **kwargs: KwargsT) -> str: ...
    def app_web_api_version(self, **kwargs: KwargsT) -> str: ...
    app_webapiVersion = app_web_api_version
    def app_build_info(self, **kwargs: KwargsT) -> BuildInfoDictionary: ...
    app_buildInfo = app_build_info
    def app_shutdown(self, **kwargs: KwargsT) -> None: ...
    def app_preferences(
        self,
        **kwargs: KwargsT,
    ) -> ApplicationPreferencesDictionary: ...
    def app_set_preferences(
        self,
        prefs: Optional[Mapping[Text, JsonValueT]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    app_setPreferences = app_set_preferences
    def app_default_save_path(self, **kwargs: KwargsT) -> str: ...
    app_defaultSavePath = app_default_save_path



from typing import Iterable
from typing import Optional
from typing import Text

from qbittorrentapi._types import DictInputT
from qbittorrentapi._types import JsonDictionaryT
from qbittorrentapi._types import KwargsT
from qbittorrentapi._types import ListInputT
from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry

class SearchJobDictionary(JsonDictionaryT):
    def __init__(
        self,
        data: DictInputT,
        client: SearchAPIMixIn,
    ) -> None: ...
    def stop(self, **kwargs: KwargsT) -> None: ...
    def status(self, **kwargs: KwargsT) -> SearchStatusesList: ...
    def results(
        self,
        limit: Optional[Text | int] = None,
        offset: Optional[Text | int] = None,
        **kwargs: KwargsT
    ) -> SearchResultsDictionary: ...
    def delete(self, **kwargs: KwargsT) -> None: ...

class SearchResultsDictionary(JsonDictionaryT): ...
class SearchStatus(ListEntry): ...

class SearchStatusesList(List[SearchStatus]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: SearchAPIMixIn,
    ) -> None: ...

class SearchCategory(ListEntry): ...

class SearchCategoriesList(List[SearchCategory]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: SearchAPIMixIn,
    ) -> None: ...

class SearchPlugin(ListEntry): ...

class SearchPluginsList(List[SearchPlugin]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: SearchAPIMixIn,
    ) -> None: ...

class Search(ClientCache):
    def start(
        self,
        pattern: Optional[Text] = None,
        plugins: Optional[Iterable[Text]] = None,
        category: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> SearchJobDictionary: ...
    def stop(
        self,
        search_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def status(
        self,
        search_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> SearchStatusesList: ...
    def results(
        self,
        search_id: Optional[Text | int] = None,
        limit: Optional[Text | int] = None,
        offset: Optional[Text | int] = None,
        **kwargs: KwargsT
    ) -> SearchResultsDictionary: ...
    def delete(
        self,
        search_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def categories(
        self,
        plugin_name: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> SearchCategoriesList: ...
    @property
    def plugins(self) -> SearchPluginsList: ...
    def install_plugin(
        self,
        sources: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    installPlugin = install_plugin
    def uninstall_plugin(
        self,
        sources: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    uninstallPlugin = uninstall_plugin
    def enable_plugin(
        self,
        plugins: Optional[Iterable[Text]] = None,
        enable: Optional[bool] = None,
        **kwargs: KwargsT
    ) -> None: ...
    enablePlugin = enable_plugin
    def update_plugins(self, **kwargs: KwargsT) -> None: ...
    updatePlugins = update_plugins

class SearchAPIMixIn(AppAPIMixIn):
    @property
    def search(self) -> Search: ...
    def search_start(
        self,
        pattern: Optional[Text] = None,
        plugins: Optional[Iterable[Text]] = None,
        category: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> SearchJobDictionary: ...
    def search_stop(
        self,
        search_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def search_status(
        self,
        search_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> SearchStatusesList: ...
    def search_results(
        self,
        search_id: Optional[Text | int] = None,
        limit: Optional[Text | int] = None,
        offset: Optional[Text | int] = None,
        **kwargs: KwargsT
    ) -> SearchResultsDictionary: ...
    def search_delete(
        self,
        search_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def search_categories(
        self,
        plugin_name: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> SearchCategoriesList: ...
    def search_plugins(self, **kwargs: KwargsT) -> SearchPluginsList: ...
    def search_install_plugin(
        self,
        sources: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    search_installPlugin = search_install_plugin
    def search_uninstall_plugin(
        self,
        names: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    search_uninstallPlugin = search_uninstall_plugin
    def search_enable_plugin(
        self,
        plugins: Optional[Iterable[Text]] = None,
        enable: Optional[bool] = None,
        **kwargs: KwargsT
    ) -> None: ...
    search_enablePlugin = search_enable_plugin
    def search_update_plugins(self, **kwargs: KwargsT) -> None: ...
    search_updatePlugins = search_update_plugins



import glob
import os
from contextlib import contextmanager
from os import environ
from os import path
from sys import path as sys_path
from time import sleep

import pytest

from qbittorrentapi import APIConnectionError
from qbittorrentapi import Client
from qbittorrentapi._version_support import (
    APP_VERSION_2_API_VERSION_MAP as api_version_map,
)
from qbittorrentapi._version_support import v
from tests.utils import CHECK_SLEEP
from tests.utils import check
from tests.utils import get_torrent
from tests.utils import mkpath
from tests.utils import retry
from tests.utils import setup_environ

QBT_VERSION, IS_QBT_DEV = setup_environ()

BASE_PATH = sys_path[0]
RESOURCES_PATH = path.join(BASE_PATH, "tests", "_resources")
assert BASE_PATH.split("/")[-1] == "qbittorrent-api"

# fmt: off
ORIG_TORRENT_URL = "https://releases.ubuntu.com/22.04.1/ubuntu-22.04.1-desktop-amd64.iso.torrent"
ORIG_TORRENT_HASH = "3b245504cf5f11bbdbe1201cea6a6bf45aee1bc0"
ORIG_TORRENT = None

TORRENT1_FILENAME = "kubuntu-22.04.2-desktop-amd64.iso.torrent"
TORRENT1_URL = "https://cdimage.ubuntu.com/kubuntu/releases/22.04/release/" + TORRENT1_FILENAME
TORRENT1_HASH = "0ee141f56407236b8acd136d56332f87674650d5"
TORRENT1_FILE_HANDLE = open(path.join(RESOURCES_PATH, TORRENT1_FILENAME), mode="rb")
TORRENT1_FILE = TORRENT1_FILE_HANDLE.read()

TORRENT2_FILENAME = "xubuntu-22.04.2-desktop-amd64.iso.torrent"
TORRENT2_URL = "https://cdimage.ubuntu.com/xubuntu/releases/22.04/release/" + TORRENT2_FILENAME
TORRENT2_HASH = "3b2dda82a16378994dbb22c49bbb91c74d2fb19c"

ROOT_FOLDER_TORRENT_FILENAME = "root_folder.torrent"
ROOT_FOLDER_TORRENT_HASH = "a14553bd936a6d496402082454a70ea7a9521adc"
ROOT_FOLDER_TORRENT_FILE_HANDLE = open(path.join(RESOURCES_PATH, ROOT_FOLDER_TORRENT_FILENAME), mode="rb")
ROOT_FOLDER_TORRENT_FILE = ROOT_FOLDER_TORRENT_FILE_HANDLE.read()
# fmt: on


@pytest.fixture(autouse=True)
def abort_if_qbittorrent_crashes(client):
    """Abort tests if qbittorrent seemingly disappears during testing."""
    try:
        client.app_version()
    except APIConnectionError:
        pytest.exit("qBittorrent crashed :(")


@pytest.fixture(autouse=True)
def skip_if_not_implemented(request, api_version):
    """Skips test if `skipif_before_api_version` marker specifies minimum API
    version."""
    if request.node.get_closest_marker("skipif_before_api_version"):
        version = request.node.get_closest_marker("skipif_before_api_version").args[0]
        if v(api_version) < v(version):
            pytest.skip("testing %s; needs %s or later" % (v(api_version), version))


@pytest.fixture(autouse=True)
def skip_if_implemented(request, api_version):
    """Skips test if `skipif_after_api_version` marker specifies maximum API version."""
    if request.node.get_closest_marker("skipif_after_api_version"):
        version = request.node.get_closest_marker("skipif_after_api_version").args[0]
        if v(api_version) >= v(version):
            pytest.skip("testing %s; needs before %s" % (v(api_version), version))


@pytest.fixture(scope="session", autouse=True)
def client():
    """QBittorrent Client for testing session."""
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERBOSE_RESPONSE_LOGGING=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    client.auth_log_in()
    # add ORIG_TORRENT to qBittorrent
    client.torrents_add(urls=ORIG_TORRENT_URL, upload_limit=10, download_limit=10)
    check(lambda: [t.hash for t in client.torrents_info()], ORIG_TORRENT_HASH, True)
    # update preferences
    client.app.preferences = dict(
        # enable RSS fetching
        rss_processing_enabled=True,
        # prevent banning IPs
        web_ui_max_auth_fail_count=1000,
        web_ui_ban_duration=1,
    )
    return client


@pytest.fixture(scope="function")
def orig_torrent(client):
    """Torrent to remain in qBittorrent for entirety of session."""
    global ORIG_TORRENT
    if ORIG_TORRENT is None:
        ORIG_TORRENT = get_torrent(client, torrent_hash=ORIG_TORRENT_HASH)
    ORIG_TORRENT.sync_local()  # ensure torrent is up-to-date
    return ORIG_TORRENT


@contextmanager
def new_torrent_standalone(client, torrent_hash=TORRENT1_HASH, **kwargs):
    check_limit = int(10 / CHECK_SLEEP)

    def add_test_torrent(torrent_hash, **kwargs):
        for attempt in range(check_limit):
            if kwargs:
                client.torrents.add(**kwargs)
            else:
                client.torrents.add(
                    torrent_files=TORRENT1_FILE,
                    save_path=path.expanduser("~/test_download/"),
                    category="test_category",
                    is_paused=True,
                    upload_limit=1024,
                    download_limit=2048,
                    is_sequential_download=True,
                    is_first_last_piece_priority=True,
                )
            try:
                return get_torrent(client, torrent_hash)
            except Exception:
                if attempt >= check_limit - 1:
                    raise
                sleep(CHECK_SLEEP)

    @retry()
    def delete_test_torrent(client, torrent_hash):
        client.torrents_delete(torrent_hashes=torrent_hash, delete_files=True)
        check(
            lambda: [t.hash for t in client.torrents_info()],
            torrent_hash,
            reverse=True,
            negate=True,
        )

    try:
        yield add_test_torrent(torrent_hash, **kwargs)
    finally:
        delete_test_torrent(client, torrent_hash)


@pytest.fixture(scope="function")
def new_torrent(client):
    """Torrent that is added on demand to qBittorrent and then removed."""
    with new_torrent_standalone(client) as torrent:
        yield torrent


@pytest.fixture(scope="session")
def app_version(client):
    """qBittorrent Version being used for testing."""
    if IS_QBT_DEV:
        return client.app.version
    return QBT_VERSION


@pytest.fixture(scope="session")
def api_version(client):
    """qBittorrent Web API Version being used for testing."""
    try:
        return api_version_map[QBT_VERSION]
    except KeyError as exp:
        if IS_QBT_DEV:
            return client.app.web_api_version
        raise exp


def pytest_sessionfinish(session, exitstatus):
    for fh in [TORRENT1_FILE_HANDLE, ROOT_FOLDER_TORRENT_FILE_HANDLE]:
        try:
            fh.close()
        except Exception:
            pass
    if environ.get("CI") != "true":
        client = Client()
        try:
            # remove all torrents
            for torrent in client.torrents_info():
                torrent.delete(delete_files=True)
        except Exception:
            pass
        # delete coverage files if not in CI
        for file in glob.iglob(path.join(BASE_PATH, ".coverage*")):
            os.unlink(file)
        # delete downloaded files if not in CI
        for filename in [TORRENT1_FILENAME, TORRENT2_FILENAME]:
            try:
                os.unlink(mkpath("~", filename))
            except Exception:
                pass



from typing import Optional
from typing import Text

from qbittorrentapi._types import JsonValueT
from qbittorrentapi._types import KwargsT
from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import Dictionary

# mypy crashes when this is imported from _types...
JsonDictionaryT = Dictionary[Text, JsonValueT]

class SyncMainDataDictionary(JsonDictionaryT): ...
class SyncTorrentPeersDictionary(JsonDictionaryT): ...

class Sync(ClientCache):
    maindata: _MainData
    torrent_peers: _TorrentPeers
    torrentPeers: _TorrentPeers
    def __init__(self, client: SyncAPIMixIn) -> None: ...

    class _MainData(ClientCache):
        _rid: int | None
        def __init__(self, client: SyncAPIMixIn) -> None: ...
        def __call__(
            self,
            rid: Optional[Text | int] = None,
            **kwargs: KwargsT,
        ) -> SyncMainDataDictionary: ...
        def delta(self, **kwargs: KwargsT) -> SyncMainDataDictionary: ...
        def reset_rid(self) -> None: ...

    class _TorrentPeers(ClientCache):
        _rid: int | None
        def __init__(self, client: SyncAPIMixIn) -> None: ...
        def __call__(
            self,
            torrent_hash: Optional[Text] = None,
            rid: Optional[Text | int] = None,
            **kwargs: KwargsT
        ) -> SyncTorrentPeersDictionary: ...
        def delta(
            self,
            torrent_hash: Optional[Text] = None,
            **kwargs: KwargsT,
        ) -> SyncTorrentPeersDictionary: ...
        def reset_rid(self) -> None: ...

class SyncAPIMixIn(AppAPIMixIn):
    @property
    def sync(self) -> Sync: ...
    def sync_maindata(
        self,
        rid: Text | int = 0,
        **kwargs: KwargsT,
    ) -> SyncMainDataDictionary: ...
    def sync_torrent_peers(
        self,
        torrent_hash: Optional[Text] = None,
        rid: Text | int = 0,
        **kwargs: KwargsT
    ) -> SyncTorrentPeersDictionary: ...
    sync_torrentPeers = sync_torrent_peers



from typing import Optional
from typing import Text

from qbittorrentapi._types import KwargsT
from qbittorrentapi._types import ListInputT
from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry

class LogPeer(ListEntry): ...

class LogPeersList(List[LogPeer]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: LogAPIMixIn,
    ) -> None: ...

class LogEntry(ListEntry): ...

class LogMainList(List[LogEntry]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: LogAPIMixIn,
    ) -> None: ...

class Log(ClientCache):
    main: _Main
    def __init__(self, client: LogAPIMixIn) -> None: ...
    def peers(
        self,
        last_known_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> LogPeersList: ...

    class _Main(ClientCache):
        def _api_call(
            self,
            normal: Optional[bool] = None,
            info: Optional[bool] = None,
            warning: Optional[bool] = None,
            critical: Optional[bool] = None,
            last_known_id: Optional[bool] = None,
            **kwargs: KwargsT
        ) -> LogMainList: ...
        def __call__(
            self,
            normal: Optional[bool] = None,
            info: Optional[bool] = None,
            warning: Optional[bool] = None,
            critical: Optional[bool] = None,
            last_known_id: Optional[bool] = None,
            **kwargs: KwargsT
        ) -> LogMainList: ...
        def info(
            self,
            last_known_id: Optional[Text | int] = None,
            **kwargs: KwargsT,
        ) -> LogMainList: ...
        def normal(
            self,
            last_known_id: Optional[Text | int] = None,
            **kwargs: KwargsT,
        ) -> LogMainList: ...
        def warning(
            self,
            last_known_id: Optional[Text | int] = None,
            **kwargs: KwargsT,
        ) -> LogMainList: ...
        def critical(
            self,
            last_known_id: Optional[Text | int] = None,
            **kwargs: KwargsT,
        ) -> LogMainList: ...

class LogAPIMixIn(AppAPIMixIn):
    @property
    def log(self) -> Log: ...
    def log_main(
        self,
        normal: Optional[bool] = None,
        info: Optional[bool] = None,
        warning: Optional[bool] = None,
        critical: Optional[bool] = None,
        last_known_id: Optional[bool] = None,
        **kwargs: KwargsT
    ) -> LogMainList: ...
    def log_peers(
        self,
        last_known_id: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> LogPeersList: ...



import pytest

from qbittorrentapi import Client
from qbittorrentapi.exceptions import APIConnectionError


def test_is_logged_in():
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    assert client.is_logged_in is False

    client.auth_log_in()
    assert client.is_logged_in is True

    client.auth_log_out()
    assert client.is_logged_in is False


def test_is_logged_in_bad_client():
    client = Client(
        host="asdf",
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    assert client.is_logged_in is False

    with pytest.raises(APIConnectionError):
        client.auth_log_in()
    assert client.is_logged_in is False

    with pytest.raises(APIConnectionError):
        client.auth_log_out()
    assert client.is_logged_in is False


def test_session_cookie(app_version):
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    assert client._session_cookie() is None

    # make the client perform a login
    assert client.app.version == app_version

    # should test other cookie names but it's difficult to change
    curr_sess_cookie = client._http_session.cookies["SID"]
    assert client._SID == curr_sess_cookie
    assert client._session_cookie() == curr_sess_cookie


def test_login_context_manager():
    with Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    ) as client:
        assert client.is_logged_in
    assert not client.is_logged_in



from typing import Any
from typing import Mapping
from typing import Optional
from typing import Text

from qbittorrentapi._types import KwargsT
from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.auth import AuthAPIMixIn
from qbittorrentapi.log import LogAPIMixIn
from qbittorrentapi.request import Request
from qbittorrentapi.rss import RSSAPIMixIn
from qbittorrentapi.search import SearchAPIMixIn
from qbittorrentapi.sync import SyncAPIMixIn
from qbittorrentapi.torrents import TorrentsAPIMixIn
from qbittorrentapi.transfer import TransferAPIMixIn

class Client(
    LogAPIMixIn,
    SyncAPIMixIn,
    TransferAPIMixIn,
    TorrentsAPIMixIn,
    RSSAPIMixIn,
    SearchAPIMixIn,
    AuthAPIMixIn,
    AppAPIMixIn,
    Request,
):
    def __init__(
        self,
        host: Text = "",
        port: Optional[Text | int] = None,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        EXTRA_HEADERS: Optional[Mapping[Text, Text]] = None,
        REQUESTS_ARGS: Optional[Mapping[Text, Any]] = None,
        VERIFY_WEBUI_CERTIFICATE: Optional[bool] = True,
        FORCE_SCHEME_FROM_HOST: Optional[bool] = False,
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS: Optional[
            bool
        ] = False,
        RAISE_ERROR_FOR_UNSUPPORTED_QBITTORRENT_VERSIONS: Optional[bool] = False,
        VERBOSE_RESPONSE_LOGGING: Optional[bool] = False,
        SIMPLE_RESPONSES: Optional[bool] = False,
        DISABLE_LOGGING_DEBUG_OUTPUT: Optional[bool] = False,
        **kwargs: KwargsT
    ) -> None: ...



from typing import Optional

from requests.exceptions import HTTPError as RequestsHTTPError
from requests.exceptions import RequestException

class APIError(Exception): ...
class UnsupportedQbittorrentVersion(APIError): ...
class FileError(IOError, APIError): ...
class TorrentFileError(FileError): ...
class TorrentFileNotFoundError(TorrentFileError): ...
class TorrentFilePermissionError(TorrentFileError): ...
class APIConnectionError(RequestException, APIError): ...
class LoginFailed(APIConnectionError): ...

class HTTPError(RequestsHTTPError, APIConnectionError):
    http_status_code: Optional[int] = None

class HTTP4XXError(HTTPError): ...
class HTTP5XXError(HTTPError): ...
class HTTP400Error(HTTP4XXError): ...
class HTTP401Error(HTTP4XXError): ...
class HTTP403Error(HTTP4XXError): ...
class HTTP404Error(HTTP4XXError): ...
class HTTP405Error(HTTP4XXError): ...
class HTTP409Error(HTTP4XXError): ...
class HTTP415Error(HTTP4XXError): ...
class HTTP500Error(HTTP5XXError): ...
class MissingRequiredParameters400Error(HTTP400Error): ...
class InvalidRequest400Error(HTTP400Error): ...
class Unauthorized401Error(HTTP401Error): ...
class Forbidden403Error(HTTP403Error): ...
class NotFound404Error(HTTP404Error): ...
class MethodNotAllowed405Error(HTTP404Error): ...
class Conflict409Error(HTTP409Error): ...
class UnsupportedMediaType415Error(HTTP415Error): ...
class InternalServerError500Error(HTTP500Error): ...



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from configparser import ConfigParser
from datetime import datetime

base_path = os.path.abspath("../..")
sys.path.insert(0, os.path.join(base_path, "src"))

setup_cfg = ConfigParser()
setup_cfg.read(os.path.join(base_path, "setup.cfg"))


# -- Project information -----------------------------------------------------
project = setup_cfg["metadata"]["name"]
copyright = "{}, {}".format(datetime.today().year, setup_cfg["metadata"]["author"])
author = setup_cfg["metadata"]["author"]

# The full version, including alpha/beta/rc tags
version = release = "v" + setup_cfg["metadata"]["version"]


# -- General configuration ---------------------------------------------------
pygments_style = "sphinx"

# warn about everything
nitpicky = True

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.githubpages",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
]

source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

linkcheck_ignore = [
    # ignore reference to just the HTTP schemas
    r"^http://$",
    r"^https://$",
    # ignore GH wiki since the check for anchors always fails
    r"https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-\(qBittorrent-4.1\)#",
    r"^http://localhost",
    r"^http://example.com",
]

# -- Options for HTML output -------------------------------------------------
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

html_theme = "furo"

# sphinx-autoapi
# extensions.append('autoapi.extension')
# autoapi_type = 'python'
# autoapi_dirs = ['../../qbittorrentapi']
# autoapi_options = ['show-inheritance-diagram']
# autoapi_ignore = ['*decorators*', '*exceptions*']

# Add mappings
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "requests": ("https://requests.readthedocs.io/en/latest/", None),
}

suppress_warnings = [
    # Suppress builds warnings while building epub
    "epub.unknown_project_files",
]

# -- Options for spelling -------------------------------------------

# Spelling check needs an additional module that is not installed by default.
# Add it only if spelling check is requested so docs can be generated without it.
if "spelling" in sys.argv:
    extensions.append("sphinxcontrib.spelling")

# Spelling language.
spelling_lang = "en_US"

# Location of word list.
spelling_word_list_filename = "spelling_wordlist"

spelling_ignore_pypi_package_names = False



from enum import Enum

try:
    from collections import UserList
except ImportError:  # pragma: no cover
    from UserList import UserList

from qbittorrentapi._attrdict import AttrDict


class APINames(Enum):
    """
    API namespaces for API endpoints.

    e.g ``torrents`` in ``http://localhost:8080/api/v2/torrents/addTrackers``
    """

    Authorization = "auth"
    Application = "app"
    Log = "log"
    Sync = "sync"
    Transfer = "transfer"
    Torrents = "torrents"
    RSS = "rss"
    Search = "search"
    EMPTY = ""


class TorrentState(Enum):
    """
    Torrent States as defined by qBittorrent.

    Definitions:
        - wiki: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-torrent-list>`_
        - code: `<https://github.com/qbittorrent/qBittorrent/blob/5dcc14153f046209f1067299494a82e5294d883a/src/base/bittorrent/torrent.h#L73>`_

    :Usage:
        >>> from qbittorrentapi import Client, TorrentState
        >>> client = Client()
        >>> # print torrent hashes for torrents that are downloading
        >>> for torrent in client.torrents_info():
        >>>     # check if torrent is downloading
        >>>     if torrent.state_enum.is_downloading:
        >>>         print(f'{torrent.hash} is downloading...')
        >>>     # the appropriate enum member can be directly derived
        >>>     state_enum = TorrentState(torrent.state)
        >>>     print(f'{torrent.hash}: {state_enum.value}')
    """

    ERROR = "error"
    MISSING_FILES = "missingFiles"
    UPLOADING = "uploading"
    PAUSED_UPLOAD = "pausedUP"
    QUEUED_UPLOAD = "queuedUP"
    STALLED_UPLOAD = "stalledUP"
    CHECKING_UPLOAD = "checkingUP"
    FORCED_UPLOAD = "forcedUP"
    ALLOCATING = "allocating"
    DOWNLOADING = "downloading"
    METADATA_DOWNLOAD = "metaDL"
    FORCED_METADATA_DOWNLOAD = "forcedMetaDL"
    PAUSED_DOWNLOAD = "pausedDL"
    QUEUED_DOWNLOAD = "queuedDL"
    FORCED_DOWNLOAD = "forcedDL"
    STALLED_DOWNLOAD = "stalledDL"
    CHECKING_DOWNLOAD = "checkingDL"
    CHECKING_RESUME_DATA = "checkingResumeData"
    MOVING = "moving"
    UNKNOWN = "unknown"

    @property
    def is_downloading(self):
        """Returns ``True`` if the State is categorized as Downloading."""
        return self in {
            TorrentState.DOWNLOADING,
            TorrentState.METADATA_DOWNLOAD,
            TorrentState.FORCED_METADATA_DOWNLOAD,
            TorrentState.STALLED_DOWNLOAD,
            TorrentState.CHECKING_DOWNLOAD,
            TorrentState.PAUSED_DOWNLOAD,
            TorrentState.QUEUED_DOWNLOAD,
            TorrentState.FORCED_DOWNLOAD,
        }

    @property
    def is_uploading(self):
        """Returns ``True`` if the State is categorized as Uploading."""
        return self in {
            TorrentState.UPLOADING,
            TorrentState.STALLED_UPLOAD,
            TorrentState.CHECKING_UPLOAD,
            TorrentState.QUEUED_UPLOAD,
            TorrentState.FORCED_UPLOAD,
        }

    @property
    def is_complete(self):
        """Returns ``True`` if the State is categorized as Complete."""
        return self in {
            TorrentState.UPLOADING,
            TorrentState.STALLED_UPLOAD,
            TorrentState.CHECKING_UPLOAD,
            TorrentState.PAUSED_UPLOAD,
            TorrentState.QUEUED_UPLOAD,
            TorrentState.FORCED_UPLOAD,
        }

    @property
    def is_checking(self):
        """Returns ``True`` if the State is categorized as Checking."""
        return self in {
            TorrentState.CHECKING_UPLOAD,
            TorrentState.CHECKING_DOWNLOAD,
            TorrentState.CHECKING_RESUME_DATA,
        }

    @property
    def is_errored(self):
        """Returns ``True`` if the State is categorized as Errored."""
        return self in {TorrentState.MISSING_FILES, TorrentState.ERROR}

    @property
    def is_paused(self):
        """Returns ``True`` if the State is categorized as Paused."""
        return self in {TorrentState.PAUSED_UPLOAD, TorrentState.PAUSED_DOWNLOAD}


TorrentStates = TorrentState


class TrackerStatus(Enum):
    """
    Tracker Statuses as defined by qBittorrent.

    Definitions:
        - wiki: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-torrent-trackers>`_
        - code: `<https://github.com/qbittorrent/qBittorrent/blob/5dcc14153f046209f1067299494a82e5294d883a/src/base/bittorrent/trackerentry.h#L42>`_

    :Usage:
        >>> from qbittorrentapi import Client, TrackerStatus
        >>> client = Client()
        >>> # print torrent hashes for torrents that are downloading
        >>> for torrent in client.torrents_info():
        >>>     for tracker in torrent.trackers:
        >>>         # display status for each tracker
        >>>         print(f"{torrent.hash[-6:]}: {TrackerStatus(tracker.status).display:>13} :{tracker.url}")
    """

    DISABLED = 0
    NOT_CONTACTED = 1
    WORKING = 2
    UPDATING = 3
    NOT_WORKING = 4

    @property
    def display(self):
        """Returns a descriptive display value for status."""
        return {
            TrackerStatus.DISABLED: "Disabled",
            TrackerStatus.NOT_CONTACTED: "Not contacted",
            TrackerStatus.WORKING: "Working",
            TrackerStatus.UPDATING: "Updating",
            TrackerStatus.NOT_WORKING: "Not working",
        }[self]


class ClientCache(object):
    """
    Caches the client.

    Subclass this for any object that needs access to the Client.
    """

    def __init__(self, *args, **kwargs):
        self._client = kwargs.pop("client")
        super(ClientCache, self).__init__(*args, **kwargs)


class Dictionary(ClientCache, AttrDict):
    """Base definition of dictionary-like objects returned from qBittorrent."""

    def __init__(self, data=None, client=None):
        super(Dictionary, self).__init__(self._normalize(data or {}), client=client)
        # allows updating properties that aren't necessarily a part of the AttrDict
        self._setattr("_allow_invalid_attributes", True)

    @classmethod
    def _normalize(cls, data):
        """Iterate through a dict converting any nested dicts to AttrDicts."""
        if isinstance(data, dict):
            return AttrDict({key: cls._normalize(value) for key, value in data.items()})
        return data


class List(ClientCache, UserList):
    """Base definition for list-like objects returned from qBittorrent."""

    def __init__(self, list_entries=None, entry_class=None, client=None):
        super(List, self).__init__(
            [
                entry_class(data=entry, client=client)
                if isinstance(entry, dict)
                else entry
                for entry in list_entries or ()
            ],
            client=client,
        )


class ListEntry(Dictionary):
    """Base definition for objects within a list returned from qBittorrent."""



from time import sleep

import pytest

from qbittorrentapi import NotFound404Error
from qbittorrentapi.search import SearchJobDictionary
from qbittorrentapi.search import SearchResultsDictionary
from qbittorrentapi.search import SearchStatusesList
from tests.utils import check
from tests.utils import get_func
from tests.utils import retry

PLUGIN_NAME = "yts"
PLUGIN_URL = "https://raw.githubusercontent.com/khensolomon/leyts/master/yts.py"


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func", ("search_update_plugins", "search.update_plugins")
)
def test_update_plugins(client, client_func):
    get_func(client, client_func)()
    check(
        lambda: any(
            entry.message.startswith("Updating plugin ")
            or entry.message == "All plugins are already up to date."
            or entry.message.endswith("content was not found at the server (404)")
            for entry in reversed(client.log.main())
        ),
        True,
    )


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func", ("search_update_plugins", "search.update_plugins")
)
def test_update_plugins_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        client.search_update_plugins()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func",
    (
        ("search_plugins", "search_enable_plugin"),
        ("search.plugins", "search.enable_plugin"),
    ),
)
def test_enable_plugin(client, client_func):
    @retry()
    def enable_plugin():
        try:
            plugins = get_func(client, client_func[0])()
        except TypeError:
            plugins = get_func(client, client_func[0])
        get_func(client, client_func[1])(
            plugins=(p["name"] for p in plugins), enable=False
        )
        check(
            lambda: (p["enabled"] for p in client.search_plugins()),
            True,
            reverse=True,
            negate=True,
        )
        get_func(client, client_func[1])(
            plugins=(p["name"] for p in plugins), enable=True
        )
        check(
            lambda: (p["enabled"] for p in client.search_plugins()),
            False,
            reverse=True,
            negate=True,
        )

    enable_plugin()


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func", ("search_enable_plugin", "search.enable_plugin")
)
def test_enable_plugin_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func",
    (
        ("search_install_plugin", "search_uninstall_plugin"),
        ("search.install_plugin", "search.uninstall_plugin"),
    ),
)
def test_install_uninstall_plugin(client, client_func):
    @retry()
    def install_plugin(client, client_func):
        get_func(client, client_func[0])(sources=PLUGIN_URL)
        check(
            lambda: (p.name for p in client.search.plugins),
            PLUGIN_NAME,
            reverse=True,
        )

    @retry()
    def uninstall_plugin(client, client_func):
        get_func(client, client_func[1])(names=PLUGIN_NAME)
        check(
            lambda: (p.name for p in client.search.plugins),
            PLUGIN_NAME,
            reverse=True,
            negate=True,
        )

    install_plugin(client, client_func)
    uninstall_plugin(client, client_func)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func",
    (
        ("search_install_plugin", "search_uninstall_plugin"),
        ("search.install_plugin", "search.uninstall_plugin"),
    ),
)
def test_install_uninstall_plugin_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func[0])()
    with pytest.raises(NotImplementedError):
        get_func(client, client_func[1])()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.skipif_after_api_version("2.6")
@pytest.mark.parametrize("client_func", ("search_categories", "search.categories"))
def test_categories(client, client_func):
    check(lambda: get_func(client, client_func)(), "All categories", reverse=True)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize("client_func", ("search_categories", "search.categories"))
def test_categories_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func",
    (
        (
            "search_start",
            "search_status",
            "search_results",
            "search_stop",
            "search_delete",
        ),
        (
            "search.start",
            "search.status",
            "search.results",
            "search.stop",
            "search.delete",
        ),
    ),
)
def test_search(client, client_func):
    job = get_func(client, client_func[0])(
        pattern="Ubuntu", plugins="enabled", category="all"
    )
    statuses = get_func(client, client_func[1])(search_id=job["id"])
    assert statuses[0]["status"] == "Running"
    assert isinstance(job, SearchJobDictionary)
    assert isinstance(statuses, SearchStatusesList)
    results = get_func(client, client_func[2])(search_id=job["id"], limit=1)
    assert isinstance(results, SearchResultsDictionary)
    results = job.results()
    assert isinstance(results, SearchResultsDictionary)
    get_func(client, client_func[3])(search_id=job["id"])
    check(
        lambda: get_func(client, client_func[1])(search_id=job["id"])[0]["status"],
        "Stopped",
    )
    get_func(client, client_func[4])(search_id=job["id"])
    statuses = get_func(client, client_func[1])()
    assert not statuses


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func",
    (
        "search_start",
        "search_status",
        "search_results",
        "search_stop",
        "search_delete",
        "search.start",
        "search.status",
        "search.results",
        "search.stop",
        "search.delete",
    ),
)
def test_search_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func", (("search_stop", "search_start"), ("search.stop", "search.start"))
)
def test_stop(client, client_func):
    job = get_func(client, client_func[1])(
        pattern="Ubuntu", plugins="enabled", category="all"
    )
    sleep(1)
    get_func(client, client_func[0])(search_id=job.id)
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")

    job = get_func(client, client_func[1])(
        pattern="Ubuntu", plugins="enabled", category="all"
    )
    sleep(1)
    job.stop()
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func", ("search_stop", "search_start", "search.stop", "search.start")
)
def test_stop_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
def test_delete(client):
    job = client.search_start(pattern="Ubuntu", plugins="enabled", category="all")
    job.delete()
    with pytest.raises(NotFound404Error):
        job.status()


@pytest.mark.skipif_after_api_version("2.1.1")
def test_delete_not_implemented(client):
    with pytest.raises(NotImplementedError):
        client.search_stop(search_id=100)



from logging import Logger
from typing import IO
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Literal
from typing import Mapping
from typing import Optional
from typing import Text
from typing import Tuple
from typing import TypeVar

from qbittorrentapi._types import DictInputT
from qbittorrentapi._types import FilesToSendT
from qbittorrentapi._types import JsonDictionaryT
from qbittorrentapi._types import KwargsT
from qbittorrentapi._types import ListInputT
from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry
from qbittorrentapi.definitions import TorrentState

logger: Logger

TorrentStatusesT = Literal[
    "all",
    "downloading",
    "seeding",
    "completed",
    "paused",
    "active",
    "inactive",
    "resumed",
    "stalled",
    "stalled_uploading",
    "stalled_downloading",
    "checking",
    "moving",
    "errored",
]

TorrentFilesT = TypeVar(
    "TorrentFilesT",
    bytes,
    Text,
    IO[bytes],
    Mapping[Text, bytes | Text | IO[bytes]],
    Iterable[bytes | Text | IO[bytes]],
)

class TorrentDictionary(JsonDictionaryT):
    def __init__(self, data: DictInputT, client: TorrentsAPIMixIn) -> None: ...
    def sync_local(self) -> None: ...
    @property
    def state_enum(self) -> TorrentState: ...
    @property
    def info(self) -> TorrentDictionary: ...
    def resume(self, **kwargs: KwargsT) -> None: ...
    def pause(self, **kwargs: KwargsT) -> None: ...
    def delete(
        self,
        delete_files: Optional[bool] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def recheck(self, **kwargs: KwargsT) -> None: ...
    def reannounce(self, **kwargs: KwargsT) -> None: ...
    def increase_priority(self, **kwargs: KwargsT) -> None: ...
    increasePrio = increase_priority
    def decrease_priority(self, **kwargs: KwargsT) -> None: ...
    decreasePrio = decrease_priority
    def top_priority(self, **kwargs: KwargsT) -> None: ...
    topPrio = top_priority
    def bottom_priority(self, **kwargs: KwargsT) -> None: ...
    bottomPrio = bottom_priority
    def set_share_limits(
        self,
        ratio_limit: Optional[Text | int] = None,
        seeding_time_limit: Optional[Text | int] = None,
        **kwargs: KwargsT
    ) -> None: ...
    setShareLimits = set_share_limits
    @property
    def download_limit(self) -> TorrentLimitsDictionary: ...
    @download_limit.setter
    def download_limit(self, v: Text | int) -> None: ...
    @property
    def downloadLimit(self) -> TorrentLimitsDictionary: ...
    @downloadLimit.setter
    def downloadLimit(self, v: Text | int) -> None: ...
    def set_download_limit(
        self,
        limit: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setDownloadLimit = set_download_limit
    @property
    def upload_limit(self) -> TorrentLimitsDictionary: ...
    @upload_limit.setter
    def upload_limit(self, v: Text | int) -> None: ...
    @property
    def uploadLimit(self) -> TorrentLimitsDictionary: ...
    @uploadLimit.setter
    def uploadLimit(self, v: Text | int) -> None: ...
    def set_upload_limit(
        self,
        limit: Optional[Text | int] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setUploadLimit = set_upload_limit
    def set_location(
        self, location: Optional[Text] = None, **kwargs: KwargsT
    ) -> None: ...
    setLocation = set_location
    def set_download_path(
        self,
        download_path: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def set_save_path(
        self,
        save_path: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setSavePath = set_save_path
    setDownloadPath = set_download_path
    def set_category(
        self,
        category: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setCategory = set_category
    def set_auto_management(
        self,
        enable: Optional[bool] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setAutoManagement = set_auto_management
    def toggle_sequential_download(self, **kwargs: KwargsT) -> None: ...
    toggleSequentialDownload = toggle_sequential_download
    def toggle_first_last_piece_priority(self, **kwargs: KwargsT) -> None: ...
    toggleFirstLastPiecePrio = toggle_first_last_piece_priority
    def set_force_start(
        self,
        enable: Optional[bool] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setForceStart = set_force_start
    def set_super_seeding(
        self,
        enable: Optional[bool] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    setSuperSeeding = set_super_seeding
    @property
    def properties(self) -> TorrentPropertiesDictionary: ...
    @property
    def trackers(self) -> TrackersList: ...
    @trackers.setter
    def trackers(self, v: Iterable[Text]) -> None: ...
    @property
    def webseeds(self) -> WebSeedsList: ...
    @property
    def files(self) -> TorrentFilesList: ...
    def rename_file(
        self,
        file_id: Optional[Text | int] = None,
        new_file_name: Optional[Text] = None,
        old_path: Optional[Text] = None,
        new_path: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    renameFile = rename_file
    def rename_folder(
        self,
        old_path: Optional[Text] = None,
        new_path: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    renameFolder = rename_folder
    @property
    def piece_states(self) -> TorrentPieceInfoList: ...
    @property
    def pieceStates(self) -> TorrentPieceInfoList: ...
    @property
    def piece_hashes(self) -> TorrentPieceInfoList: ...
    @property
    def pieceHashes(self) -> TorrentPieceInfoList: ...
    def add_trackers(
        self,
        urls: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    addTrackers = add_trackers
    def edit_tracker(
        self,
        orig_url: Optional[Text] = None,
        new_url: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    editTracker = edit_tracker
    def remove_trackers(
        self,
        urls: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    removeTrackers = remove_trackers
    def file_priority(
        self,
        file_ids: Optional[int | Iterable[Text | int]] = None,
        priority: Optional[Text | int] = None,
        **kwargs: KwargsT
    ) -> None: ...
    filePriority = file_priority
    def rename(self, new_name: Optional[Text] = None, **kwargs: KwargsT) -> None: ...
    def add_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    addTags = add_tags
    def remove_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    removeTags = remove_tags
    def export(self, **kwargs: KwargsT) -> bytes: ...

class TorrentPropertiesDictionary(JsonDictionaryT): ...
class TorrentLimitsDictionary(JsonDictionaryT): ...
class TorrentCategoriesDictionary(JsonDictionaryT): ...
class TorrentsAddPeersDictionary(JsonDictionaryT): ...
class TorrentFile(ListEntry): ...

class TorrentFilesList(List[TorrentFile]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: TorrentsAPIMixIn,
    ) -> None: ...

class WebSeed(ListEntry): ...

class WebSeedsList(List[WebSeed]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: TorrentsAPIMixIn,
    ) -> None: ...

class Tracker(ListEntry): ...

class TrackersList(List[Tracker]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: TorrentsAPIMixIn,
    ) -> None: ...

class TorrentInfoList(List[TorrentDictionary]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: TorrentsAPIMixIn,
    ) -> None: ...

class TorrentPieceData(ListEntry): ...

class TorrentPieceInfoList(List[TorrentPieceData]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: TorrentsAPIMixIn,
    ) -> None: ...

class Tag(ListEntry): ...

class TagList(List[Tag]):
    def __init__(
        self,
        list_entries: ListInputT,
        client: TorrentsAPIMixIn,
    ) -> None: ...

class Torrents(ClientCache):
    info: _Info
    resume: _ActionForAllTorrents
    pause: _ActionForAllTorrents
    delete: _ActionForAllTorrents
    recheck: _ActionForAllTorrents
    reannounce: _ActionForAllTorrents
    increase_priority: _ActionForAllTorrents
    increasePrio: _ActionForAllTorrents
    decrease_priority: _ActionForAllTorrents
    decreasePrio: _ActionForAllTorrents
    top_priority: _ActionForAllTorrents
    topPrio: _ActionForAllTorrents
    bottom_priority: _ActionForAllTorrents
    bottomPrio: _ActionForAllTorrents
    download_limit: _ActionForAllTorrents
    downloadLimit: _ActionForAllTorrents
    upload_limit: _ActionForAllTorrents
    uploadLimit: _ActionForAllTorrents
    set_download_limit: _ActionForAllTorrents
    setDownloadLimit: _ActionForAllTorrents
    set_share_limits: _ActionForAllTorrents
    setShareLimits: _ActionForAllTorrents
    set_upload_limit: _ActionForAllTorrents
    setUploadLimit: _ActionForAllTorrents
    set_location: _ActionForAllTorrents
    setLocation: _ActionForAllTorrents
    set_save_path: _ActionForAllTorrents
    setSavePath: _ActionForAllTorrents
    set_download_path: _ActionForAllTorrents
    setDownloadPath: _ActionForAllTorrents
    set_category: _ActionForAllTorrents
    setCategory: _ActionForAllTorrents
    set_auto_management: _ActionForAllTorrents
    setAutoManagement: _ActionForAllTorrents
    toggle_sequential_download: _ActionForAllTorrents
    toggleSequentialDownload: _ActionForAllTorrents
    toggle_first_last_piece_priority: _ActionForAllTorrents
    toggleFirstLastPiecePrio: _ActionForAllTorrents
    set_force_start: _ActionForAllTorrents
    setForceStart: _ActionForAllTorrents
    set_super_seeding: _ActionForAllTorrents
    setSuperSeeding: _ActionForAllTorrents
    add_peers: _ActionForAllTorrents
    addPeers: _ActionForAllTorrents
    def __init__(self, client: TorrentsAPIMixIn) -> None: ...
    def add(
        self,
        urls: Optional[Iterable[Text]] = None,
        torrent_files: Optional[TorrentFilesT] = None,
        save_path: Optional[Text] = None,
        cookie: Optional[Text] = None,
        category: Optional[Text] = None,
        is_skip_checking: Optional[bool] = None,
        is_paused: Optional[bool] = None,
        is_root_folder: Optional[bool] = None,
        rename: Optional[Text] = None,
        upload_limit: Optional[Text | int] = None,
        download_limit: Optional[Text | int] = None,
        use_auto_torrent_management: Optional[bool] = None,
        is_sequential_download: Optional[bool] = None,
        is_first_last_piece_priority: Optional[bool] = None,
        tags: Optional[Iterable[Text]] = None,
        content_layout: Optional[
            Literal["Original", "Subfolder", "NoSubFolder"]
        ] = None,
        ratio_limit: Optional[Text | float] = None,
        seeding_time_limit: Optional[Text | int] = None,
        download_path: Optional[Text] = None,
        use_download_path: Optional[bool] = None,
        stop_condition: Optional[Literal["MetadataReceived", "FilesChecked"]] = None,
        **kwargs: KwargsT
    ) -> Text: ...

    class _ActionForAllTorrents(ClientCache):
        func: Callable[..., Any]
        def __init__(
            self,
            client: TorrentsAPIMixIn,
            func: Callable[..., Any],
        ) -> None: ...
        def __call__(
            self,
            torrent_hashes: Optional[Iterable[Text]] = None,
            **kwargs: KwargsT,
        ) -> Optional[Any]: ...
        def all(self, **kwargs: KwargsT) -> Optional[Any]: ...

    class _Info(ClientCache):
        def __call__(
            self,
            status_filter: Optional[TorrentStatusesT] = None,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def all(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def downloading(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def seeding(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def completed(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def paused(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def active(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def inactive(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def resumed(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def stalled(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def stalled_uploading(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def stalled_downloading(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def checking(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def moving(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...
        def errored(
            self,
            category: Optional[Text] = None,
            sort: Optional[Text] = None,
            reverse: Optional[bool] = None,
            limit: Optional[Text | int] = None,
            offset: Optional[Text | int] = None,
            torrent_hashes: Optional[Iterable[Text]] = None,
            tag: Optional[Text] = None,
            **kwargs: KwargsT
        ) -> TorrentInfoList: ...

class TorrentCategories(ClientCache):
    @property
    def categories(self) -> TorrentCategoriesDictionary: ...
    @categories.setter
    def categories(self, v: Iterable[Text]) -> None: ...
    def create_category(
        self,
        name: Optional[Text] = None,
        save_path: Optional[Text] = None,
        download_path: Optional[Text] = None,
        enable_download_path: Optional[bool] = None,
        **kwargs: KwargsT
    ) -> None: ...
    createCategory = create_category
    def edit_category(
        self,
        name: Optional[Text] = None,
        save_path: Optional[Text] = None,
        download_path: Optional[Text] = None,
        enable_download_path: Optional[bool] = None,
        **kwargs: KwargsT
    ) -> None: ...
    editCategory = edit_category
    def remove_categories(
        self,
        categories: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    removeCategories = remove_categories

class TorrentTags(ClientCache):
    @property
    def tags(self) -> TagList: ...
    @tags.setter
    def tags(self, v: Optional[Iterable[Text]] = None) -> None: ...
    def add_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    addTags = add_tags
    def remove_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    removeTags = remove_tags
    def create_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    createTags = create_tags
    def delete_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    deleteTags = delete_tags

class TorrentsAPIMixIn(AppAPIMixIn):
    @property
    def torrents(self) -> Torrents: ...
    @property
    def torrent_categories(self) -> TorrentCategories: ...
    @property
    def torrent_tags(self) -> TorrentTags: ...
    def torrents_add(
        self,
        urls: Optional[Iterable[Text]] = None,
        torrent_files: Optional[TorrentFilesT] = None,
        save_path: Optional[Text] = None,
        cookie: Optional[Text] = None,
        category: Optional[Text] = None,
        is_skip_checking: Optional[bool] = None,
        is_paused: Optional[bool] = None,
        is_root_folder: Optional[bool] = None,
        rename: Optional[Text] = None,
        upload_limit: Optional[Text | int] = None,
        download_limit: Optional[Text | int] = None,
        use_auto_torrent_management: Optional[bool] = None,
        is_sequential_download: Optional[bool] = None,
        is_first_last_piece_priority: Optional[bool] = None,
        tags: Optional[Iterable[Text]] = None,
        content_layout: Optional[
            Literal["Original", "Subfolder", "NoSubFolder"]
        ] = None,
        ratio_limit: Optional[Text | float] = None,
        seeding_time_limit: Optional[Text | int] = None,
        download_path: Optional[Text] = None,
        use_download_path: Optional[bool] = None,
        stop_condition: Optional[Literal["MetadataReceived", "FilesChecked"]] = None,
        **kwargs: KwargsT
    ) -> Literal["Ok.", "Fails."]: ...
    @staticmethod
    def _normalize_torrent_files(
        user_files: TorrentFilesT,
    ) -> Tuple[FilesToSendT, list[IO[bytes]]] | Tuple[None, None]: ...
    def torrents_properties(
        self,
        torrent_hash: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> TorrentPropertiesDictionary: ...
    def torrents_trackers(
        self,
        torrent_hash: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> TrackersList: ...
    def torrents_webseeds(
        self,
        torrent_hash: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> WebSeedsList: ...
    def torrents_files(
        self,
        torrent_hash: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> TorrentFilesList: ...
    def torrents_piece_states(
        self,
        torrent_hash: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> TorrentPieceInfoList: ...
    torrents_pieceStates = torrents_piece_states
    def torrents_piece_hashes(
        self,
        torrent_hash: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> TorrentPieceInfoList: ...
    torrents_pieceHashes = torrents_piece_hashes
    def torrents_add_trackers(
        self,
        torrent_hash: Optional[Text] = None,
        urls: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_addTrackers = torrents_add_trackers
    def torrents_edit_tracker(
        self,
        torrent_hash: Optional[Text] = None,
        original_url: Optional[Text] = None,
        new_url: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_editTracker = torrents_edit_tracker
    def torrents_remove_trackers(
        self,
        torrent_hash: Optional[Text] = None,
        urls: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_removeTrackers = torrents_remove_trackers
    def torrents_file_priority(
        self,
        torrent_hash: Optional[Text] = None,
        file_ids: Optional[int | Iterable[Text | int]] = None,
        priority: Optional[Text | int] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_filePrio = torrents_file_priority
    def torrents_rename(
        self,
        torrent_hash: Optional[Text] = None,
        new_torrent_name: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    def torrents_rename_file(
        self,
        torrent_hash: Optional[Text] = None,
        file_id: Optional[Text | int] = None,
        new_file_name: Optional[Text] = None,
        old_path: Optional[Text] = None,
        new_path: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_renameFile = torrents_rename_file
    def torrents_rename_folder(
        self,
        torrent_hash: Optional[Text] = None,
        old_path: Optional[Text] = None,
        new_path: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_renameFolder = torrents_rename_folder
    def torrents_export(
        self,
        torrent_hash: Optional[Text] = None,
        **kwargs: KwargsT,
    ) -> bytes: ...
    def torrents_info(
        self,
        status_filter: Optional[TorrentStatusesT] = None,
        category: Optional[Text] = None,
        sort: Optional[Text] = None,
        reverse: Optional[bool] = None,
        limit: Optional[Text | int] = None,
        offset: Optional[Text | int] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        tag: Optional[Text] = None,
        **kwargs: KwargsT
    ) -> TorrentInfoList: ...
    def torrents_resume(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def torrents_pause(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def torrents_delete(
        self,
        delete_files: bool = False,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    def torrents_recheck(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def torrents_reannounce(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    def torrents_increase_priority(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_increasePrio = torrents_increase_priority
    def torrents_decrease_priority(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_decreasePrio = torrents_decrease_priority
    def torrents_top_priority(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_topPrio = torrents_top_priority
    def torrents_bottom_priority(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_bottomPrio = torrents_bottom_priority
    def torrents_download_limit(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> TorrentLimitsDictionary: ...
    torrents_downloadLimit = torrents_download_limit
    def torrents_set_download_limit(
        self,
        limit: Optional[Text | int] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setDownloadLimit = torrents_set_download_limit
    def torrents_set_share_limits(
        self,
        ratio_limit: Optional[Text | int] = None,
        seeding_time_limit: Optional[Text | int] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setShareLimits = torrents_set_share_limits
    def torrents_upload_limit(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> TorrentLimitsDictionary: ...
    torrents_uploadLimit = torrents_upload_limit
    def torrents_set_upload_limit(
        self,
        limit: Optional[Text | int] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setUploadLimit = torrents_set_upload_limit
    def torrents_set_location(
        self,
        location: Optional[Text] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setLocation = torrents_set_location
    def torrents_set_save_path(
        self,
        save_path: Optional[Text] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setSavePath = torrents_set_save_path
    def torrents_set_download_path(
        self,
        download_path: Optional[Text] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setDownloadPath = torrents_set_download_path
    def torrents_set_category(
        self,
        category: Optional[Text] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setCategory = torrents_set_category
    def torrents_set_auto_management(
        self,
        enable: Optional[bool] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setAutoManagement = torrents_set_auto_management
    def torrents_toggle_sequential_download(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_toggleSequentialDownload = torrents_toggle_sequential_download
    def torrents_toggle_first_last_piece_priority(
        self,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_toggleFirstLastPiecePrio = torrents_toggle_first_last_piece_priority
    def torrents_set_force_start(
        self,
        enable: Optional[bool] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setForceStart = torrents_set_force_start
    def torrents_set_super_seeding(
        self,
        enable: Optional[bool] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_setSuperSeeding = torrents_set_super_seeding
    def torrents_add_peers(
        self,
        peers: Optional[Iterable[Text]] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> TorrentsAddPeersDictionary: ...
    torrents_addPeers = torrents_add_peers
    def torrents_categories(self, **kwargs: KwargsT) -> TorrentCategoriesDictionary: ...
    def torrents_create_category(
        self,
        name: Optional[Text] = None,
        save_path: Optional[Text] = None,
        download_path: Optional[Text] = None,
        enable_download_path: Optional[bool] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_createCategory = torrents_create_category
    def torrents_edit_category(
        self,
        name: Optional[Text] = None,
        save_path: Optional[Text] = None,
        download_path: Optional[Text] = None,
        enable_download_path: Optional[bool] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_editCategory = torrents_edit_category
    def torrents_remove_categories(
        self,
        categories: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_removeCategories = torrents_remove_categories
    def torrents_tags(self, **kwargs: KwargsT) -> TagList: ...
    def torrents_add_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_addTags = torrents_add_tags
    def torrents_remove_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        torrent_hashes: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT
    ) -> None: ...
    torrents_removeTags = torrents_remove_tags
    def torrents_create_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_createTags = torrents_create_tags
    def torrents_delete_tags(
        self,
        tags: Optional[Iterable[Text]] = None,
        **kwargs: KwargsT,
    ) -> None: ...
    torrents_deleteTags = torrents_delete_tags



from time import sleep

import pytest

from qbittorrentapi._version_support import v
from qbittorrentapi.exceptions import APIError
from qbittorrentapi.rss import RSSitemsDictionary
from tests.utils import check
from tests.utils import get_func
from tests.utils import retry

FOLDER_ONE = "testFolderOne"
FOLDER_TWO = "testFolderTwo"

ITEM_ONE = "YTS"
ITEM_TWO = "YTSNew"
URL = "https://yts.mx/rss/0/all/all/0/en"


@pytest.fixture(scope="function", autouse=True)
def rss_feed(client, api_version):
    def delete_feed(name):
        try:
            client.rss_remove_item(item_path=name)
            check(lambda: client.rss_items(), name, reverse=True, negate=True)
        except Exception:
            pass

    if v(api_version) >= v("2.2"):
        client.app.preferences = dict(rss_auto_downloading_enabled=False)
        # refreshing the feed is finicky...so try several times if necessary
        done = False
        for i in range(5):
            delete_feed(ITEM_ONE)
            client.rss.add_feed(url=URL, item_path=ITEM_ONE)
            check(lambda: client.rss_items(), ITEM_ONE, reverse=True)
            # wait until feed is refreshed
            for j in range(20):
                if client.rss.items.with_data[ITEM_ONE]["articles"]:
                    done = True
                    yield ITEM_ONE
                    delete_feed(ITEM_ONE)
                    break
                sleep(0.25)
            if done:
                break
        else:
            raise Exception("RSS Feed '%s' did not refresh..." % ITEM_ONE)
    else:
        yield ""


@pytest.mark.skipif_before_api_version("2.2.1")
@pytest.mark.parametrize(
    "client_func",
    ["rss_refresh_item", "rss_refreshItem", "rss.refresh_item", "rss.refreshItem"],
)
def test_refresh_item(client, rss_feed, client_func):
    @retry()
    def run_test():
        get_func(client, client_func)(item_path=rss_feed)
        check(
            lambda: client.rss_items(include_feed_data=True)[rss_feed]["lastBuildDate"],
            "",
            negate=True,
        )
        last_refresh = client.rss_items(include_feed_data=True)[rss_feed][
            "lastBuildDate"
        ]
        sleep(1)
        get_func(client, client_func)(item_path=rss_feed)
        check(
            lambda: client.rss_items(include_feed_data=True)[rss_feed]["lastBuildDate"],
            last_refresh,
            negate=True,
        )

    run_test()


# inconsistent behavior with endpoint for API version 2.2
@pytest.mark.skipif_after_api_version("2.2")
@pytest.mark.parametrize(
    "client_func",
    ["rss_refresh_item", "rss_refreshItem", "rss.refresh_item", "rss.refreshItem"],
)
def test_refresh_item_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize("client_func", ["rss_items", "rss.items"])
def test_items(client, rss_feed, client_func):
    check(lambda: get_func(client, client_func)(), rss_feed, reverse=True)
    check(
        lambda: get_func(client, client_func)(include_feed_data=True),
        rss_feed,
        reverse=True,
    )
    check(
        lambda: get_func(client, client_func)(include_feed_data=True)[rss_feed],
        "articles",
        reverse=True,
    )

    if "." in client_func:
        check(
            lambda: get_func(client, client_func).without_data,
            rss_feed,
            reverse=True,
        )
        check(
            lambda: get_func(client, client_func).with_data[rss_feed],
            "articles",
            reverse=True,
        )


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize("client_func", ["rss_items", "rss.items"])
def test_add_feed(client, rss_feed, client_func):
    assert rss_feed in get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.9.1")
@pytest.mark.parametrize(
    "client_func",
    ["rss_set_feed_url", "rss_setFeedURL", "rss.set_feed_url", "rss.setFeedURL"],
)
def test_set_feed_url(client, rss_feed, client_func):
    curr_feed_url = client.rss_items()[rss_feed].url
    new_feed_url = curr_feed_url + "asdf"
    get_func(client, client_func)(url=new_feed_url, item_path=rss_feed)
    assert new_feed_url == client.rss_items()[rss_feed].url


@pytest.mark.skipif_after_api_version("2.9.1")
@pytest.mark.parametrize(
    "client_func",
    ["rss_set_feed_url", "rss_setFeedURL", "rss.set_feed_url", "rss.setFeedURL"],
)
def test_set_feed_url_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize(
    "client_func",
    ["rss_remove_item", "rss_removeItem", "rss.remove_item", "rss.removeItem"],
)
def test_remove_feed(client, rss_feed, client_func):
    get_func(client, client_func)(item_path=rss_feed)
    check(lambda: client.rss_items(), rss_feed, reverse=True, negate=True)


@pytest.mark.parametrize(
    "client_func",
    [
        ("rss_add_folder", "rss_remove_item"),
        ("rss_addFolder", "rss_removeItem"),
        ("rss.add_folder", "rss.remove_item"),
        ("rss.addFolder", "rss.removeItem"),
    ],
)
def test_add_remove_folder(client, client_func):
    name = "test_isos"

    get_func(client, client_func[0])(folder_path=name)  # add_folder
    check(lambda: client.rss_items(), name, reverse=True)
    get_func(client, client_func[1])(item_path=name)  # remove_folder
    check(lambda: client.rss_items(), name, reverse=True, negate=True)


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize(
    "client_func", ["rss_move_item", "rss_moveItem", "rss.move_item", "rss.moveItem"]
)
def test_move(client, rss_feed, client_func):
    try:
        new_name = "new_loc"
        get_func(client, client_func)(orig_item_path=rss_feed, new_item_path=new_name)
        check(lambda: client.rss_items(), new_name, reverse=True)
    finally:
        try:
            client.rss_remove_item(item_path=new_name)
        except APIError:
            pass


@pytest.mark.skipif_before_api_version("2.5.1")
@pytest.mark.parametrize(
    "client_func",
    ["rss_mark_as_read", "rss_markAsRead", "rss.mark_as_read", "rss.markAsRead"],
)
def test_mark_as_read(client, rss_feed, client_func):
    item_id = client.rss.items.with_data[rss_feed]["articles"][0]["id"]
    get_func(client, client_func)(item_path=rss_feed, article_id=item_id)
    check(
        lambda: client.rss.items.with_data[rss_feed]["articles"][0],
        "isRead",
        reverse=True,
    )


@pytest.mark.skipif_after_api_version("2.5.1")
@pytest.mark.parametrize(
    "client_func",
    ["rss_mark_as_read", "rss_markAsRead", "rss.mark_as_read", "rss.markAsRead"],
)
def test_mark_as_read_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize(
    "client_func",
    (
        (
            "rss_add_feed",
            "rss_set_rule",
            "rss_rules",
            "rss_rename_rule",
            "rss_matching_articles",
            "rss_remove_rule",
            "rss_remove_item",
        ),
        (
            "rss_addFeed",
            "rss_setRule",
            "rss_rules",
            "rss_renameRule",
            "rss_matchingArticles",
            "rss_removeRule",
            "rss_removeItem",
        ),
        (
            "rss.add_feed",
            "rss.set_rule",
            "rss.rules",
            "rss.rename_rule",
            "rss.matching_articles",
            "rss.remove_rule",
            "rss.remove_item",
        ),
        (
            "rss.addFeed",
            "rss.setRule",
            "rss.rules",
            "rss.renameRule",
            "rss.matchingArticles",
            "rss.removeRule",
            "rss.removeItem",
        ),
    ),
)
def test_rules(client, api_version, client_func):
    def check_for_rule(name):
        try:
            get_func(client, client_func[2])()  # rss_rules
            check(
                lambda: get_func(client, client_func[2])(), name, reverse=True
            )  # rss_rules
        except TypeError:
            check(
                lambda: get_func(client, client_func[2]), name, reverse=True
            )  # rss_rules

    rule_name = ITEM_ONE + "Rule"
    rule_name_new = rule_name + "New"
    rule_def = {"enabled": True, "affectedFeeds": URL, "addPaused": True}
    try:
        # rss_set_rule
        get_func(client, client_func[1])(rule_name=rule_name, rule_def=rule_def)
        check_for_rule(rule_name)

        if v(api_version) >= v("2.6"):  # rename was broken for a bit
            get_func(client, client_func[3])(
                orig_rule_name=rule_name, new_rule_name=rule_name_new
            )  # rss_rename_rule
            check_for_rule(rule_name_new)
        if v(api_version) >= v("2.5.1"):
            assert isinstance(
                get_func(client, client_func[4])(rule_name=rule_name),
                RSSitemsDictionary,
            )  # rss_matching_articles
        else:
            with pytest.raises(NotImplementedError):
                get_func(client, client_func[4])(
                    rule_name=rule_name
                )  # rss_matching_articles
    finally:
        get_func(client, client_func[5])(rule_name=rule_name)  # rss_remove_rule
        get_func(client, client_func[5])(rule_name=rule_name_new)  # rss_remove_rule
        check(lambda: client.rss_rules(), rule_name, reverse=True, negate=True)
        get_func(client, client_func[6])(item_path=ITEM_ONE)  # rss_remove_item
        assert ITEM_TWO not in client.rss_items()
        check(lambda: client.rss_items(), ITEM_TWO, reverse=True, negate=True)


@pytest.mark.skipif_after_api_version("2.2")
@pytest.mark.parametrize(
    "client_func",
    [
        "rss_matching_articles",
        "rss_matchingArticles",
        "rss.matching_articles",
        "rss.matchingArticles",
    ],
)
def test_rules_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        get_func(client, client_func)()
